/***************************************************************************

PROJECT:			CS 297 Deliverable 3

FILE:				tps.cpp

PURPOSE:			Implementation of the thin-plate-spline method 
					for point-set alignment.

COMPILER / TOOL:	Visual C++ .NET

PROGRAMMER:			Wallun Chan

START DATE:			4/1/05

REMARKS:			This code is based on the following:

					F. L. Bookstein. Principal warps: 
					thin-plate splines and decomposition of deformations. 
					IEEE Trans. Pattern Analysis and Machine Intelligence, 
					11(6):567585, June 1989,

					and

					Gianluca Donato and Serge Belongie. Approximation 
					Methods for Thin Plate Spline Mappings and 
					Principal Warps. October 2002;
					http://www-cse.ucsd.edu/~sjb/pami_tps.pdf.

					and

					Jarno Elonen. Thin Plate Spline editor - 
					An example program in C++. 
					http://elonen.iki.fi/code/tpsdemo, 2003.

***************************************************************************/

#include "stdafx.h"
#include "tps.h"

TPSSolver::TPSSolver()
// Default constructor
{
}

TPSSolver::~TPSSolver()
// Destructor
{
}

void TPSSolver::setPoints(const VecIntPr& trgtPts, const VecIntPr& testPts)
/* PURPOSE:		This function takes a set of test points and maps it to 
				the set of target points.  The function implements the 
				thin-plate-spline algorithm that uses the notion of a 
				thin plate "tacked onto" a set of control points and uses 
				the bending of the surface defined by splines on the control 
				points as a way of describing the mapping of one set of points 
				to another.  The function solves a matrix formula for a set of 
				weights and factors to provide a closed form equation mapping 
				the positions of one set of points to another.

   RECEIVES:	Set of test and target points such that the test points are
				mapped onto the target points.
*/
{
	int p = (int)testPts.size();
	lMatrix.SetSize(p + 3, p + 3);
	int i;
	int j;

	// Set K matrix
	for (i = 0; i < p; i++)
	{	IntPr pt1 = testPts[i];
		for (j = 0; j < p; j++)
		{	if (i != j)
			{	IntPr pt2 = testPts[j];
				double diffX = pt1.first - pt2.first;
				double diffY = pt1.second - pt2.second;
				lMatrix(i, j) = shapeFunction(diffX * diffX + diffY * diffY);
			}
			else
			{	lMatrix(i, j) = 0.0;
			}
		}
	}

	// Set P matrix and its transpose
	for (i = 0; i < p; i++)
	{	IntPr pt = testPts[i];
		lMatrix(i, p) = 1.0;
		lMatrix(i, p + 1) = pt.first;
		lMatrix(i, p + 2) = pt.second;
		lMatrix(p, i) = 1.0;
		lMatrix(p + 1, i) = pt.first;
		lMatrix(p + 2, i) = pt.second;
	}

	// Set O matrix
	for (i = p; i < p + 3; i++)
	{	lMatrix(i, p) = 0.0;
		lMatrix(i, p + 1) = 0.0;
		lMatrix(i, p + 2) = 0.0;
	}

	// Set v column vectors
	Matrix yMatrix(p + 3, 2);
    for (i = 0; i < p; i++)
    {	IntPr targPt = trgtPts[i];
		IntPr testPt = testPts[i];
		yMatrix(i, 0) = targPt.first - testPt.first;
		yMatrix(i, 1) = targPt.second - testPt.second;
    }

	// Set o column vectors
    yMatrix(p, 0) = yMatrix(p + 1, 0) = yMatrix(p + 2, 0) = 0.0;
    yMatrix(p, 1) = yMatrix(p + 1, 1) = yMatrix(p + 2, 1) = 0.0;

	// Solve for w and a coefficients
	wMatrix = lMatrix.Solve(yMatrix);
	samples = testPts;
}

void TPSSolver::mapPoints(VecIntPr& pts)
/* PURPOSE:		This function maps a set of points to the test
				image space by using the mapping function
				determined by the function setPoints().

   RECEIVES:	Set of points to be mapped test image space.
*/
{
	int n = (int)pts.size();
	int p = (int)samples.size();
	int i;
	int j;
	for (i = 0; i < n; i++)
	{	double x = pts[i].first;
		double y = pts[i].second;

		// dx = a1x + a2x * x + a3x * y
		// dy = a1y + a2y * x + a3y * y
		double dx = wMatrix(p, 0) + wMatrix(p + 1, 0) * x + wMatrix(p + 2, 0) * y;
		double dy = wMatrix(p, 1) + wMatrix(p + 1, 1) * x + wMatrix(p + 2, 1) * y;
		for (j = 0; j < p; j++)
		{	double diffX = samples[j].first - x;
			double diffY = samples[j].second - y;
			double u = shapeFunction(diffX * diffX + diffY * diffY);
			dx += wMatrix(j, 0) * u;
			dy += wMatrix(j, 1) * u;
		}
		pts[i].first += (int)dx;
		pts[i].second += (int)dy;
	}
}

double TPSSolver::shapeFunction(double rad)
/* PURPOSE:		To calculate the shap

   RECEIVES:	Radial distance between two points.

   REMARKS:		Input is passed in as a square since don't want to
				perform a square root, given that the equation is
				shapeFunction(rad) = rad ^ 2 * log(rad).
*/
{
	// log(r ^ 2) = 2 * log(r)
	return (rad == 0.0) ? 0.0 : rad * log(rad) / 2.0;

}