/***************************************************************************

PROJECT:			CS 297 Deliverable 3

FILE:				utility.cpp

PURPOSE:			Implements various functions to support random number 
					generation, mapping points to shape context histograms, 
					storing point coordinates, setting up the shape context 
					for a set of points defining a shape, obtaining the edges 
					from an image via edge detection, overlaying points and 
					edges to an image, sampling pixels from edges, and parsing 
					a .pgm image file.

COMPILER / TOOL:	Visual C++ .NET

PROGRAMMER:			Wallun Chan

START DATE:			4/1/05

***************************************************************************/ 

#include "stdafx.h"
#include "utility.h"
#include "lap.h"
#include "tps.h"

const char* DELIMITERS = " \t\r\n";
const char* P5_FORMAT = "P5";

// Global random number generator
Random random;

Random::Random()
// Initialize random number generator
{ 
	srand((unsigned)time(0));
}

double Random::randDouble()
/* PURPOSE:		To obtain random double.

   RETURNS:		Random double n in 0.0 < n < 1.0.
*/
{ 
	return (double)rand() / RAND_MAX;
}

int Random::randRange(int min, int max)
/* PURPOSE:		To obtain number between min and max.

   RECEIVES:	min and max integers specifying range
				of random integer to return

   RETURNS:		Random integer from min to max, exclusive
				of max.
*/
{ 
	return min + (int)(randDouble() * (max - min));
}

/**************************************************************************************/

void meanStdDevRadDist(const VecIntPr& pts, const IntPr& center, 
double& mean, double& stdDev)
/* PURPOSE:		To calculate the centroid, and mean and standard
				deviation of the radial distances between the
				group centroid and all points.

   RECEIVES:	Points with (x, y) coordinates, centroid coordinate,
				and mean and standard deviation of radial distance
				between group centroid to be calculated.
*/
{
	/* Calculate mean radial distance between centroid and
	   points. */
	mean = 0.0;
	int i;
	int n = (int)pts.size();
	for (i = 0; i < n; i++)
	{	double difX = pts[i].first - center.first;
		double difY = pts[i].second - center.second;
		mean += sqrt(difX * difX + difY * difY);
	}
	mean /= n;

	/* Calculate standard deviation of radial distance 
	   between centroid and points. */
	stdDev = 0.0;
	for (i = 0; i < n; i++)
	{	double difX = pts[i].first - center.first;
		double difY = pts[i].second - center.second;
		double dist = sqrt(difX * difX + difY * difY);
		double dif = mean - dist;
		stdDev += (dif * dif);
	}
	stdDev = sqrt(stdDev / (n - 1));
}

IntPr centroid(const VecIntPr& pts)
/* PURPOSE:		Calculate centroid of point set.

   RECEIVES:	Points with (x, y) coordinates.

   RETURNS:		Centroid in (x, y) coordinate.
*/
{
	int sumX = 0;
	int sumY = 0;
	int n = (int)pts.size();
	int i;
	for (i = 0; i < n; i++)
	{	sumX += pts[i].first;
		sumY += pts[i].second;
	}
	sumX = (int)((double)sumX / n);
	sumY = (int)((double)sumY / n);
	return IntPr(sumX, sumY);
}

VecIntPr removeOutliers(const VecIntPr& pts, int threshold)
/* PURPOSE:		Takes a set of points and attempts to remove
				outliers with respect to the mean and standard
				deviation of the radial distances between the
				group centroid and all points.

   RECEIVES:	Points with (x, y) coordinates and factor
				specifying the number of standard deviations
				away from the mean.

   RETURNS:		Original set of points with all outliers with 
				radial distances greater than the specified number 
				of standard deviations from the mean are removed.
*/
{
	double meanRadDist;
	double stdDevRadDist;
	IntPr center = centroid(pts);
	meanStdDevRadDist(pts, center, meanRadDist, stdDevRadDist);
	double metric = meanRadDist + stdDevRadDist * threshold;
	VecIntPr adjPts;
	int i;
	for (i = 0; i < (int)pts.size(); i++)
	{	double difX = pts[i].first - center.first;
		double difY = pts[i].second - center.second;
		double dist = sqrt(difX * difX + difY * difY);
		if (dist < metric)
		{	adjPts.push_back(pts[i]);
		}
	}
	return adjPts;
}

void minMaxMeanRadDist(const VecIntPr& pts, double& min, 
double& max, double& mean)
/* PURPOSE:		Calculates the min, max, and mean radial 
				distances between all pairs of points on 
				a shape.

   RECEIVES:	Points in (x, y) coordinates, and min, max,
				and mean radial distances to be calculated.
*/
{
	min = LARGE;
	max = 0.0;
	mean = 0.0;
	int i;
	int j;
	int num = (int)pts.size();
	for (i = 0; i < num; i++)
	{	IntPr pt1 = pts[i];
		for (j = 0; j < num; j++)
		{	if (i != j)
			{	IntPr pt2 = pts[j];
				double xDif = pt1.first - pt2.first;
				double yDif = pt1.second - pt2.second;
				double dist = sqrt(xDif * xDif + yDif * yDif);
				if (dist < min)
				{	min = dist;
				}
				if (dist > max)
				{	max = dist;
				}
				mean += dist;
			}
		}
	}
	mean /= (num * (num - 1));
}

void shapeContextParams(const VecIntPr& tempPts, const VecIntPr& testPts, 
int radialBinNum, int angBinNum, double& radialBinSize, double& angBinSize, 
double& scale, double& tempMean, double& testMean)
/* PURPOSE:		Sets the size of the log-polar histogram
				in the angular and radial directions.

   RECEIVES:	Sampled points for the template and test images, number
				of radial bins, number of angular bins, sizes of the
				radial bin and angular bin to be calculated, scale
				factor to be multiplied with the template image to make
				it the same size as the test image, and mean radial
				distances of the template and test shapes to be 
				calculated.
*/
{
	double tempMin;
	double tempMax;
	double testMin;
	double testMax;
	minMaxMeanRadDist(tempPts, tempMin, tempMax, tempMean);
	minMaxMeanRadDist(testPts, testMin, testMax, testMean);

	/* Scale is equal to the reciprocal of the minimum 
	   normalized radial distance. */
	scale = 1.0 / __min(tempMin / tempMean, testMin / testMean);

	// Maximum normalized radial distance
	double maxDist = __max(tempMax / tempMean, testMax / testMean);

	// Split logarithmic radial distance into equal parts
	radialBinSize = log(scale * maxDist) / radialBinNum;

	// Size of angular bin
	angBinSize = 360.0 / angBinNum;
}

void directionVectors(const VecIntPr& pts, DirectionMap& dirMap)
/* PURPOSE:		Calculates a rotation invariant axis to measure the 
				angular orientation for each sampled point for a
				given shape.

   RECEIVES:	Points with (x, y) coordinates and map object that
				associates a point with its rotation invariant
				angular measurement axis.
*/
{
	int i;
	int j;
	for (i = 0; i < (int)pts.size(); i++)
	{	double sumX = 0.0;
		double sumY = 0.0;
		for (j = 0; j < (int)pts.size(); j++)
		{	if (i != j)
			{	sumX += (pts[i].first - pts[j].first);
				sumY += (pts[i].second - pts[j].second);
			}
		}
		double angle = atan2(sumY, sumX) * 180.0 / PI;
		angle = angle < 0.0 ? 360 + angle : angle;
		dirMap.insert(PrDouble(pts[i], angle));
	}
}

void setShapeContext(Map& shapeContext, const VecIntPr& pts, 
int radialBinNum, int angBinNum, double radialBinSize, 
double angBinSize, double scale, double mean)
/* PURPOSE:		Calculates the shape context for a shape by
				determining the histogram for each sampled
				point of a shape.

   RECEIVES:	Map object that maps point coordinates with
				the associated shape context histogram of
				a point.
*/
{
	DirectionMap dirMap;
	directionVectors(pts, dirMap);
	int i;
	int j;
	for (i = 0; i < (int)pts.size(); i++)
	{	IntPr pt1 = pts[i];

		// Set size for log-polar histogram
		MatInt* histogram = new MatInt(radialBinNum, angBinNum);
		histogram->Null();

		/* Calculate rotation invariant axis relative to horizontal axis 
		   to measure angular orientation of point. */
		directionIter dirIter = dirMap.find(pt1);

		// Sort points in their respective log-polar histograms
		for (j = 0; j < (int)pts.size(); j++)
		{	if (i != j)
			{	IntPr pt2 = pts[j];
				double xDif = pt1.first - pt2.first;
				double yDif = pt1.second - pt2.second;

				// Calculate log radial distance
				double logDist = log(scale * sqrt(xDif * xDif + yDif * yDif) / mean);

				// Sort in radial direction
				int radialBinIndex = (int)(logDist / radialBinSize);
				if (radialBinIndex == radialBinNum)
				{	radialBinIndex--;
				}

				// Calculate angular orientation
				double angle = atan2(yDif, xDif) * 180.0 / PI;
				angle = (angle < 0.0) ? 360.0 + angle : angle;
				angle = (angle > dirIter->second) ? angle - 
					dirIter->second : 360.0 - (dirIter->second - angle);

				// Sort in angular direction
				int angBinIndex = (int)(angle / angBinSize);
				if (angBinIndex == angBinNum)
				{	angBinIndex--;
				}

				// Place point in bin
				(*histogram)(radialBinIndex, angBinIndex)++;
			}
		}
		// Map point coordinate with log-polar histogram
		shapeContext.insert(PrMat(pt1, histogram));
	}
}

void setShapeContextPair(Map& tempContext, Map& testContext, 
const VecIntPr& tempSampledPts, const VecIntPr& testSampledPts, 
int radialBinNum, int angBinNum)
/* PURPOSE:		Determine shape context for all sampled points
				of template and test images.

   RECEIVES:	Map objects for template and test images that
				associates point coordinates with corresponding
				log-polar histogram for each sampled point,
				sampled points of template and test images,
				and number of radial and angular bins.
*/
{
	double tempMean;
	double testMean;
	double scale;	
	double radialBinSize;
	double angBinSize;

	// Obtain log-polar histogram sizing data
	shapeContextParams(tempSampledPts, testSampledPts, radialBinNum,
		angBinNum, radialBinSize, angBinSize, scale, tempMean, testMean);

	// Sets shape context for sampled points of template image
	setShapeContext(tempContext, tempSampledPts, radialBinNum, 
		angBinNum, radialBinSize, angBinSize, scale, tempMean);

	// Sets shape context for sampled points of test image
	setShapeContext(testContext, testSampledPts, radialBinNum, 
		angBinNum, radialBinSize, angBinSize, scale, testMean);
}

/**************************************************************************************/

BOOL getSampledPoints(VecIntPr& sampledPts, BYTE* edgeMap, 
int width, int height, int sampleNum)
/* PURPOSE:		Samples points of an edge detected image.

   RECEIVES:	List to store sampled points, edge detected
				image, width and height of edge detected
				image, and number of points to sample.

   RETURNS:		True if sampling was successful, and false
				otherwise.
*/
{
	VecIntPr tmp;
	int i;
	int j;
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
		{	if ((unsigned)edgeMap[i * width + j] > 0)
			{	tmp.push_back(IntPr(j, i));
			}
		}
	}
	if (sampleNum > (int)tmp.size())
	{	return FALSE;
	}
	for (i = 0; i < sampleNum; i++)
	{	int num = random.randRange(0, (int)tmp.size());
		sampledPts.push_back(tmp[num]);

		// Don't erase an otherwise empty vector
		if (!tmp.empty() && num < (int)tmp.size())
		{	tmp.erase(tmp.begin() + num);
		}
	}
	return TRUE;
}

double cost(const Map& tempShape, const Map& testShape, 
const IntPr& tempPt, const IntPr& testPt)
/* PURPOSE:		Calculates the chi-squared distance between
				points on two shapes.

   RECEIVES:	Shape context mapping of template and test
				images, and coordinates of points to
				calculate the associated chi-squared 
				distance.

   RETURNS:		Cost of assigning a particular template
				and test sampled point pair.
*/
{
	MapIter tempIt = tempShape.find(tempPt);
	MapIter testIt = testShape.find(testPt);
	MatInt* tempHistogram = (*tempIt).second;
	MatInt* testHistogram = (*testIt).second;
	double sum = 0.0;
	int i;
	int j;
	for (i = 0; i < (int)tempHistogram->RowNo(); i++)
	{	for (j = 0; j < (int)tempHistogram->ColNo(); j++)
		{	int tempBin = (*tempHistogram)(i, j);
			int testBin = (*testHistogram)(i, j);
			if (tempBin + testBin != 0)
			{	double subtract = tempBin - testBin;
				double add = tempBin + testBin;
				sum += (subtract * subtract / add);
			}
		}
	}
	return sum / 2.0;
}

void matchPoints(const Map& tempShape, const Map& testShape, 
const VecIntPr& tempPts, VecIntPr& testPts)
/* PURPOSE:		Takes sampled point-sets from the edge detected template
				and test images, and based on the calculated shape
				contexts for each set, rearranges the order so that
				a point-to-point correspondence between the sets is
				established.

   RECEIVES:	Shape context mapping of template and test images,
				and sampled point-sets of template and test images.

   REMARKS:		The shape context mapping contains an associated
				histogram for each point coordinate.
*/
{
	int dim = (int)tempPts.size();

	// Set shape context cost matrix
	MatInt mat(dim, dim);
	int i;
	int j;
	for (i = 0; i < dim; i++)
	{	for (j = 0; j < dim; j++)
		{	mat(i, j) = (int)cost(tempShape, testShape, tempPts[i], testPts[j]);
		}
	}
	VecInt rowSol(dim);
	VecInt colSol(dim);
	VecInt u(dim);
	VecInt v(dim);

	/* Call algorithm to solve assignment problem to find 
	   best matching between points. */
	lap(dim, mat, rowSol, colSol, u, v);

	/* Order sampled test image points with respect to matching found by
	   Hungarian algorithm. */
	VecIntPr tmp;
	for (i = 0; i < dim; i++)
	{	tmp.push_back(testPts[rowSol[i]]);
	}
	testPts = tmp;
}

VecIntPr getPointList(BYTE* image, int width, int height)
/* PURPOSE:		Takes an edge detected image and retrieves the
				point coordinates of non-zero pixels, and stores
				the coordinates in a vector.

   RECEIVES:	Edge detected image, and width and height of image.

   RETURNS:		List with coordinates of edge detected pixels.
*/
{
	VecIntPr pts;
	int i;
	int j;
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
		{	if (image[i * width + j] != 0)
			{	pts.push_back(IntPr(j, i));
			}
		}
	}
	return pts;
}

/**************************************************************************************/

DWORD* rgbFormat(BYTE* image, int byteCount)
/* PURPOSE:		Takes grayscale image data and converts it into 
				a form suitable to be displayed in a compatible 
				device context.

   RECEIVES:	Grayscale image data and byte count of image data.

   RETURNS:		Grayscale image converted into a four-byte group
				format.

   REMARKS:		Specifically, the first byte of the four-byte group is
				reserved, the three bytes that follow are for red,
				green, and blue values.
*/
{
	DWORD* bytes = new DWORD[byteCount];
	int i;
	for (i = 0; i < byteCount; i++)
	{	DWORD pixel = (DWORD)image[i];
		bytes[i] = pixel << 16 | pixel << 8 | pixel & 0x00FFFFFF;
	}
	return bytes;
}

DWORD* rgbFormat(BYTE* image, int byteCount, COLORREF color)
/* PURPOSE:		Takes edge detected image data and converts it into 
				a form suitable to be displayed by a compatible 
				device context.

   RECEIVES:	Edge detected image data, byte count of image data,
				and color of edge data to be displayed.

   RETURNS:		Edge detected image converted into a four-byte group
				format.

   REMARKS:		Specifically, the first byte of the four-byte group is
				reserved, the three bytes that follow are for red,
				green, and blue values.
*/
{
	DWORD red = GetRValue(color);
	DWORD green = GetGValue(color);
	DWORD blue = GetBValue(color);
	DWORD* bytes = new DWORD[byteCount];
	int i;
	for (i = 0; i < byteCount; i++)
	{	if ((unsigned)image[i] > 0)
		{	bytes[i] = red << 16 | green << 8 | blue & 0x00FFFFFF;
		}
		else
		{	bytes[i] = 0;
		}
	}
	return bytes;
}

DWORD* rgbFormat(VecIntPr& pts, int width, int height, COLORREF color)
/* PURPOSE:		Takes a set of points and paints them on a blank image.

   RECEIVES:	Points in (x, y) coordinates, width and height of image
				to display points, and color of displayed points.

   RETURNS:		Image with sampled points displayed on a blank image
				in the specified color.
*/
{
	DWORD red = GetRValue(color);
	DWORD green = GetGValue(color);
	DWORD blue = GetBValue(color);
	DWORD* bytes = new DWORD[width * height];
	int i;
	for (i = 0; i < width * height; i++)
	{	bytes[i] = 0;
	}
	for (i = 0; i < (int)pts.size(); i++)
	{	IntPr pt = pts[i];
		int index = pt.second * width + pt.first;
		if (index >= 0 && index < width * height)
		{	bytes[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			if (pt.first > 0 && pt.first < width - 1 && 
					pt.second > 0 && pt.second < height - 1)
			{	index = (pt.second + 1) * width + pt.first + 1;
				bytes[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second + 1) * width + pt.first - 1;
				bytes[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first + 1;
				bytes[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first - 1;
				bytes[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			}
		}
	}
	return bytes;
}

BYTE* rgbSaveFormat(BYTE* image, int width, int height)
/* PURPOSE:		Converts image data into an rgb triplet format for every
				three bytes suitable for storage as a bitmap file.

   RECEIVES:	Grayscale image intensity data, and width and height
				of image to be saved to a file.

   RETURNS:		Byte array suitable for storage as a bitmap file.
*/
{
	long extWidth = width * 3 + width % 4;
	long imageSize = extWidth * height;
	BYTE* bytes = new BYTE[imageSize];
	DWORD* imDat = (DWORD*)image;
	int i;
	int j;
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
		{	DWORD pixel = imDat[i * width + j];
			int pIndex = i * extWidth + j * 3;
			bytes[pIndex] = (BYTE)(pixel & 0x000000FF);
			bytes[pIndex + 1] = (BYTE)((pixel & 0x0000FF00) >> 8);
			bytes[pIndex + 2] = (BYTE)((pixel & 0x00FF0000) >> 16);
		}
	}
	return bytes;
}

BYTE* overlay(BYTE* image, BYTE* edgeMap, int width, int height, 
VecIntPr& pts, COLORREF color1, COLORREF color2)
/* PURPOSE:		Overlays original grayscale image, edge detected image,
				and sampled points.

   RECEIVES:	Original grayscale image, edge detected image of grayscale
				image, width and height of images, sampled points to be
				imposed, and color of edge data and sampled points to be
				displayed.

   RETURNS:		Image with original grayscale image, edge detected image, 
				and sampled points all super-imposed.
*/
{
	DWORD red = GetRValue(color1);
	DWORD green = GetGValue(color1);
	DWORD blue = GetBValue(color1);
	DWORD* overlay = rgbFormat(image, width * height);
	int i;
	for (i = 0; i < width * height; i++)
	{	if ((unsigned)edgeMap[i] > 0)
		{	overlay[i] = red << 16 | green << 8 | blue & 0x00FFFFFF;
		}
	}
	red = GetRValue(color2);
	green = GetGValue(color2);
	blue = GetBValue(color2);
	for (i = 0; i < (int)pts.size(); i++)
	{	IntPr pt = pts[i];
		int index = pt.second * width + pt.first;
		if (index >= 0 && index < width * height)
		{	overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			if (pt.first > 0 && pt.first < width - 1 && 
					pt.second > 0 && pt.second < height - 1)
			{	index = (pt.second + 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second + 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			}
		}
	}
	return (BYTE*)overlay;
}

BYTE* overlay(BYTE* image, BYTE* edgeMap, int byteCount, COLORREF color)
/* PURPOSE:		Overlays original grayscale image with edge detected image.

   RECEIVES:	Original grayscale image, edge detected image of grayscale
				image, byte count of images, and color of edge data to be
				displayed.

   RETURNS:		Image with original grayscale image and 
				imposed edge detected image.
*/
{
	DWORD red = GetRValue(color);
	DWORD green = GetGValue(color);
	DWORD blue = GetBValue(color);
	DWORD* overlay = rgbFormat(image, byteCount);
	int i;
	for (i = 0; i < byteCount; i++)
	{	if ((unsigned)edgeMap[i] > 0)
		{	overlay[i] = red << 16 | green << 8 | blue & 0x00FFFFFF;
		}
	}
	return (BYTE*)overlay;
}

BYTE* overlay(BYTE* image, int width, int height,
VecIntPr& pts, COLORREF color)
/* PURPOSE:		Overlays original grayscale image with
				sampled points.

   RECEIVES:	Original grayscale image, width and height
				of image, sampled points to be imposed on
				original image, and color of sampled points
				to be displayed.

   RETURNS:		Image with original grayscale image and 
				imposed sampled points.
*/
{
	DWORD red = GetRValue(color);
	DWORD green = GetGValue(color);
	DWORD blue = GetBValue(color);
	DWORD* overlay = rgbFormat(image, width * height);
	int i;
	for (i = 0; i < (int)pts.size(); i++)
	{	IntPr pt = pts[i];
		int index = pt.second * width + pt.first;
		if (index >= 0 && index < width * height)
		{	overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			if (pt.first > 0 && pt.first < width - 1 && 
					pt.second > 0 && pt.second < height - 1)
			{	index = (pt.second + 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second + 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			}
		}
	}
	return (BYTE*)overlay;
}

BYTE* overlay(BYTE* image, int width, int height, VecIntPr& pts, 
COLORREF color1, COLORREF color2)
/* PURPOSE:		Overlays an edge detected image with sampled points.

   RECEIVES:	Edge detected image, width and height of image,
				sampled points to be drawn onto edge detected
				data, and colors of edge data and sampled points
				to be displayed.

   RETURNS:		Image with edge detected data and imposed sampled points.
*/
{
	DWORD red = GetRValue(color1);
	DWORD green = GetGValue(color1);
	DWORD blue = GetBValue(color1);
	DWORD* overlay = rgbFormat(image, width * height);
	int i;
	for (i = 0; i < width * height; i++)
	{	if ((unsigned)image[i] > 0)
		{	overlay[i] = red << 16 | green << 8 | blue & 0x00FFFFFF;
		}
	}
	red = GetRValue(color2);
	green = GetGValue(color2);
	blue = GetBValue(color2);
	for (i = 0; i < (int)pts.size(); i++)
	{	IntPr pt = pts[i];
		int index = pt.second * width + pt.first;
		if (index >= 0 && index < width * height)
		{	overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			if (pt.first > 0 && pt.first < width - 1 && 
					pt.second > 0 && pt.second < height - 1)
			{	index = (pt.second + 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second + 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first + 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
				index = (pt.second - 1) * width + pt.first - 1;
				overlay[index] = red << 16 | green << 8 | blue & 0x00FFFFFF;
			}
		}
	}
	return (BYTE*)overlay;
}

/**************************************************************************************/

BYTE* getEdgeMap(EdgeDetector* detector, int hysCheck,
int thinCheck, double sigma, int* threshold)
/* PURPOSE:		(Re)generates image data for user-selections
				of hysteresis and thinning, and user inputs
				for sigma and threshold.

   RECEIVES:	Edge detector containing image edge data,
				user-selections of hysteresis and thinning,
				and values for sigma (increases the amount
				of smoothing) and threshold (sensitivity
				to intensity changes).

   RETURNS:		Generated image based on user-selections of
				hysteresis and thin checkboxes, and user
				inputs for sigma and threshold.

   REMARKS:		Hysteresis is post-postprocessing of edge
				detection where image pixels that should 
				be and have not been suppressed are done
				so. Thinning is as it sounds, that is,
				edges are thinned to its minimum, i.e.
				single pixel width edges.
*/
{
	BYTE* edgeMap;
	if (hysCheck && thinCheck)
	{	
		detector->image_find_edges(sigma, threshold);
		detector->image_hysteresis(*threshold / 3);
		edgeMap = detector->image_thin();
	}
	else if (hysCheck)
	{	detector->image_find_edges(sigma, threshold);
		edgeMap = detector->image_hysteresis(*threshold / 3);
	}
	else if (thinCheck)
	{	detector->image_find_edges(sigma, threshold);
		edgeMap = detector->image_thin();
	}
	else
	{	edgeMap = detector->image_find_edges(sigma, threshold);
	}
	return edgeMap;
}

void setDisplayImage(BYTE*& displayImage, BYTE* image, BYTE* edgeMap,
VecIntPr& sampledPts, int width, int height, int imageCheck, 
int edgeCheck, int sampleCheck)
/* PURPOSE:		Depending on user-selections of the checkboxes,
				this function generates the appropriate image
				to be displayed.

   RECEIVES:	Image to generate and display, 
*/
{
	// Overlay original image, edge detected image, and sampled points
	if (imageCheck && edgeCheck && sampleCheck)
	{	displayImage = overlay(image, edgeMap, width, height, 
			sampledPts, BLUE, YLW);
	}
	// Overlay original and edge detected image
	else if (imageCheck && edgeCheck)
	{	displayImage = overlay(image, edgeMap, width * height, BLUE);
	}
	// Overly edge detected image and sampled points
	else if (edgeCheck && sampleCheck)
	{	displayImage = overlay(edgeMap, width, height, 
			sampledPts, BLUE, YLW);
	}
	// Overly original image and sampled points
	else if (imageCheck && sampleCheck)
	{	displayImage = overlay(image, width, height, sampledPts, YLW);
	}
	// Show only original image
	else if (imageCheck)
	{	displayImage = (BYTE*)rgbFormat(image, width * height);
	}
	// Show only edge data
	else if (edgeCheck)
	{	displayImage = (BYTE*)rgbFormat(edgeMap, width * height, BLUE);
	}
	// Show only sampled points
	else if (sampleCheck)
	{	displayImage = (BYTE*)rgbFormat(sampledPts, width, height, YLW);
	}
	// Show nothing
	else
	{	displayImage = 0;
	}
}

/**************************************************************************************/

BYTE* pgmByteRead(const char* file, int& width, int& height)
/* PURPOSE:		Parses a .pgm image file for image dimensions and pixel
				intensities, and stores values in unsigned char array.

   RECEIVES:	Image file path name, unsigned char array to store image
				values to, and image pixel width and height.

   RETURNS:		unsigned char array with initialized values or 0.

   REMARKS:		1) This function reads both the "P2" and "P5" version 
				   of the .pgm format; for P2, each pixel intensity
				   value is represented as an ASCII decimal number
				   string (e.g. 101).  For P5, each pixel is
				   represented in 8 or 16-bit binary.
				
				2) The number of grayscale bits is 8, giving intensity 
				   values from 0 to 255.
*/
{
	int curPos = 0;
	int index = 0;
	int linePos = 0;
	BYTE* bytes;
	CString line;
	CString token;
	CString format;
	CStdioFile fileHandle;
	if (!fileHandle.Open(file, CFile::modeRead | CFile::typeBinary)) 
	{	return 0;
	}
	while(fileHandle.ReadString(line))
	{	// Ignore comments marked by a '#'
		if (line.Find('#') == -1)
		{	// Read .pgm type (P2 or P5)
			if (linePos == 0)
			{	format = line.Tokenize(DELIMITERS, curPos);
			}
			// Read image dimensions
			else if (linePos == 1)
			{	curPos = 0;
				token = line.Tokenize(DELIMITERS, curPos);
				width = atoi(token);
				token = line.Tokenize(DELIMITERS, curPos);
				height = atoi(token);
				bytes = new BYTE[width * height];
			}
			// Check if it's P5 format, since P2 is default .pgm format
			else if (linePos == 2)
			{	// Read pixel intensities for P5 format
				if (format == P5_FORMAT)
				{	BYTE* longText = new BYTE[width * height];
					int numRead = fileHandle.Read(longText, width * height);
					for (index = 0; index < numRead; index++)
					{	bytes[index] = longText[index];
					}
					delete [] longText;
					break;
				}
			}
			// Read pixel intensities for P2 format
			else
			{	curPos = 0;
				token = line.Tokenize(DELIMITERS, curPos);
				while (token != "")
				{	bytes[index] = atoi(token);
					index++;
					token = line.Tokenize(DELIMITERS, curPos);
				}
			}
			linePos++;
		}
	}
	fileHandle.Close();
	return bytes;
}

void imageOutput(const char* file, BYTE* image, int width, int height)
/* PURPOSE:		Saves a bitmap image

   RECEIVES:	File name of bitmap file, pixel data of image, and pixel
				dimensions of image.

   REMARKS:		This is a rather crude and inflexible way of saving
				exclusively .bmp files.  But I wanted to save 
				images with the least amount of code, and this
				is probably it.
*/
{
	BITMAPINFOHEADER bmih;
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biBitCount = 24;
	bmih.biPlanes = 1;
	bmih.biCompression = BI_RGB;
	bmih.biWidth = width;
	bmih.biHeight = -height;
	bmih.biSizeImage = (width * 3 + width % 4) * height;

    BITMAPFILEHEADER bmfh;
    int nBitsOffset = sizeof(BITMAPFILEHEADER) + bmih.biSize; 
    LONG imSize = bmih.biSizeImage;
    LONG lFileSize = nBitsOffset + imSize;
    bmfh.bfType = 'B' + ('M' << 8);
    bmfh.bfOffBits = nBitsOffset;
    bmfh.bfSize = lFileSize;
    bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;

	FILE* pFile = fopen(file, "wb");
    fwrite(&bmfh, 1, sizeof(BITMAPFILEHEADER), pFile);
    fwrite(&bmih, 1, sizeof(BITMAPINFOHEADER), pFile);
	fwrite(rgbSaveFormat(image, width, height), 1, imSize, pFile);
    fclose(pFile);
}