/***************************************************************************

PROJECT:			CS 297 Deliverable 3

FILE:				utility.h

PURPOSE:			Provides various data structures and function 
					declarations to support random number generation, 
					mapping points to shape context histograms, storing 
					point coordinates, setting up the shape context for 
					a set of points defining a shape, obtaining the edges 
					from an image via edge detection, overlaying points 
					and edges to an image, sampling pixels from edges, 
					and parsing a .pgm image file.

COMPILER / TOOL:	Visual C++ .NET

PROGRAMMER:			Wallun Chan

START DATE:			4/1/05

***************************************************************************/ 

#ifndef _UTILITY_H_
#define _UTILITY_H_

#include "edgefinder.h"
#include "matrx.h"
#include <time.h>

class TPSSolver;

// Really small number
const double SMALL = 1.0E-6;

// Really big number
const double LARGE = 1E9;

// Approximation for pi constant
const double PI = 3.141592653589793238462643383279;

// Size of cross-hairs used in displaying group centroids
const int CROSS_HAIR_SIZE = 7;

// Colors to be used for drawing in device contexts
const COLORREF WHT = RGB(255, 255, 255);
const COLORREF RED = RGB(255, 0, 0);
const COLORREF GRN = RGB(0, 255, 0);
const COLORREF BLUE = RGB(0, 0, 255);
const COLORREF MAG = RGB(255, 0, 255);
const COLORREF CYAN = RGB(0, 255, 255);
const COLORREF YLW = RGB(255, 255, 0);
const COLORREF BLK = RGB(0, 0, 0);
const COLORREF GRAY = RGB(100, 100, 100);

// Border height of gray title bar in each splitter window
const int BORDER_HEIGHT = 25;

/* Used to paint entire client area black for each splitter
   window. */
const int BORDER_LIMIT = 1500;

// Default scroll length
const int DEFAULT_SCROLL = 100;

// Used to size scroll length accordingly
const double SCROLL_FACTOR = 1.25;

// Used to filter for .pgm files in "open" and "save" dialogs
static char pgmFilter[] = "PGM Files (*.pgm)|*.pgm||";
static char bmpFilter[] = "BMP Files (*.bmp)|*.bmp||";

// Integer tuple
typedef pair<int, int> IntPr;

// Tuple to store mapping between a point and its histogram
typedef pair<IntPr, MatInt*> PrMat;

// Used to map a point with its shape context histogram
typedef map<IntPr, MatInt*> Map;	

typedef Map::const_iterator MapIter;
typedef vector<int> VecInt;
typedef vector<double> VecD;
typedef vector<IntPr> VecIntPr;

// Used to locate the endpoints of a rectangle
typedef pair<IntPr, IntPr> Box;

// Stores a number of rectangle coordinates
typedef vector<Box> Boxes;

// Stores a number of point groupings
typedef vector<VecIntPr> PtGroup;

typedef pair<IntPr, double> PrDouble;

/* Used to store average vector heading for each point
   on a shape. */
typedef map<IntPr, double> DirectionMap;
typedef DirectionMap::iterator directionIter;

// Random number generator
class Random
{
public:
	Random();
	double randDouble();
	int randRange(int min, int max);
};

IntPr centroid(const VecIntPr& pts);

VecIntPr removeOutliers(const VecIntPr& pts, int threshold = 0);

void minMaxMeanRadDist(const VecIntPr& pts, double& min, 
double& max, double& mean);

void setShapeContextPair(Map& tempContext, Map& testContext, 
const VecIntPr& tempSampledPts, const VecIntPr& testSampledPts, 
int radialBinNum, int angBinNum);

BOOL getSampledPoints(VecIntPr& sampledPts, BYTE* edgeMap, 
int width, int height, int sampleNum);

void matchPoints(const Map& tempShape, const Map& testShape, 
const VecIntPr& tempPts, VecIntPr& testPts);

VecIntPr getPointList(BYTE* image, int width, int height);

BYTE* getEdgeMap(EdgeDetector* detector, int hysCheck,
int thinCheck, double sigma, int* threshold);

void setDisplayImage(BYTE*& displayImage, BYTE* image, BYTE* edgeMap,
VecIntPr& sampledPts, int width, int height, int imageCheck, 
int edgeCheck, int sampleCheck);

DWORD* rgbFormat(BYTE* image, int byteCount);

BYTE* overlay(BYTE* image, int width, int height, VecIntPr& pts,
COLORREF color1, COLORREF color2);

BYTE* rgbSaveFormat(BYTE* image, int width, int height);

BYTE* pgmByteRead(const char* file, int& width, int& height);

void imageOutput(const char* file, BYTE* image, int width, int height);

#endif